import numpy as np
import matplotlib.pyplot as plt
import time

import scp_utility as scp_ut
import bin_sch as BS
import multiArmedBandit as MAB



tf = "S2"
bf = "Standard"

### Bin scheme
def S_2(sol):
    trans = np.divide(1, ( 1 + np.exp(-1*sol) ) )
    return trans

def standar_bin(trans):
    matrixRand = np.random.uniform(low=0.0,high=1.0,size=trans.shape)
    bin = np.greater(trans,matrixRand).astype(int)

    return bin

def bin_scheme(sol, bsol, solRanking, tf, bf): 
    bs = BS.DiscretizationScheme(sol, bsol, solRanking, tf, bf)
    bin_sol = bs.binariza()
    if np.sum(bin_sol) == 0:
        print(tf)
        print(bf)
        print(sol)
        print(np.unique(bsol))
        print(solRanking)
        #exit(0)

    return np.array(bin_sol)

def update_bs(agent):
    tf, bf = agent.select_arm()
    return tf, bf

### Reparación simple
def reparaSimple(SCP, solution):
    sol =  np.reshape(solution, (SCP["n"],))
    set = SCP["S"]
    numRep = 0
    indices = list(range(set.shape[0]))
    np.random.shuffle(indices)
    for i in indices:               # i in m (filas) (200)
        if np.sum(set[i,:] * sol) == 0:                     #
            idxRestriccion = np.argwhere((set[i,:]) > 0)    # inidices de los elementos que cubren a la zona i
            a = np.argmin(SCP["C_s"][idxRestriccion])       # a: elemento con menor costo que cubre i
            idxMenorPeso = idxRestriccion[a]
            sol[idxMenorPeso[0]] = 1
            numRep += 1
    return sol

### inicializar población
def inicializar_BWO(SCP,npop,bounds):
    dof = SCP["n"]
    X = np.zeros((npop,dof))

    for i in range(npop):
        X[i,:] = np.random.uniform(bounds[0],bounds[1], size=(1,dof))
    
    return X

### Black widow
## Cross-over
def crossover(scp, pop, bpop, i1, i2, tf, bf):
    p1 = pop[i1,:]
    p2 = pop[i2,:]
    bp1 = bpop[i1,:]
    bp2 = bpop[i2,:]

    nvar = p1.shape[0]
    nvar = int(nvar/2)*2

    children = np.zeros((nvar,nvar))
    bchildren = np.zeros((nvar,nvar))
    

    # Optimizar
    for i in range(1, int(nvar/2)+1):
        alpha = np.random.uniform(low=0.0,high=1.0, size=p1.shape)
        c1  = np.multiply(p1, alpha) + np.multiply(p2, 1-alpha)
        c2  = np.multiply(p2, alpha) + np.multiply(p1, 1-alpha)
        bc1 = np.multiply(bp1, alpha) + np.multiply(bp2, 1-alpha)
        bc2 = np.multiply(bp2, alpha) + np.multiply(bp1, 1-alpha)

        children[i*2-2,:] = c1
        children[i*2-1,:] = c2

        bchildren[i*2-2,:] = reparaSimple(scp, bc1)
        bchildren[i*2-1,:] = reparaSimple(scp, bc2)

    bchildren = bin_scheme(children, bchildren, scp_ut.fitness(scp,bpop), tf, bf)
    for i in range(0, nvar):
        bchildren[i,:] = reparaSimple(scp, bchildren[i,:])


    return children, bchildren

## Canibalismo de padre
def father_cannibalism(scp, bpop1, pop1, i1, i2):
    func = scp_ut.fitness
    bp1 = bpop1[i1,:]
    bp2 = bpop1[i2,:]

    if func(scp,bp1) > func(scp,bp2): 
        pop1 = np.delete(pop1, i1, axis=0)
        bpop1 = np.delete(bpop1, i1, axis=0)
    else: 
        pop1 = np.delete(pop1, i2, axis=0)
        bpop1 = np.delete(bpop1, i2, axis=0)

    return pop1, bpop1

## Canibalismo de hermanos
def sibling_canibalism(scp, children, bchildren, cr):

    # Evaluar y ordenar hijos
    children, bchildren = ordenar_fitness(scp, children, bchildren)

    # canibalismo de hijos
    children = children[: max(int(children.shape[0] * cr), 1),:]
    bchildren = bchildren[: max(int(children.shape[0] * cr), 1),:]

    return children, bchildren

# Mutación
def mutation(scp, pop1, bpop1, nm, tf, bf):

    nvar = pop1.shape[1]
    pop3 = np.zeros((nm,nvar))
    bpop3 = np.zeros((nm,nvar))
    for i in range(0, nm):
        index = np.random.randint(0, bpop1.shape[0])
        m = pop1[index,:]
        bm = bpop1[index,:]

        cp1, cp2 = np.random.randint(0, nvar), np.random.randint(0, nvar)

        m[cp1], m[cp2] = m[cp2], m[cp1]
        bm[cp1], bm[cp2] = bm[cp2], bm[cp1]
        
        pop3[i,:] = m
        bpop3[i,:] = reparaSimple(scp, bm)
    
    bpop3 = bin_scheme(pop3, bpop3, scp_ut.fitness(scp,bpop3), tf, bf)
    for i in range(0, nm):
        bpop3[i,:] = reparaSimple(scp, bpop3[i,:])


    return pop3, bpop3

# Ordenar por fitness
def ordenar_fitness(scp, pop, bpop):
    npop = pop.shape[0]
    fitness = scp_ut.fitness(scp, bpop)
    order = np.argsort(fitness, axis=0)
    pop = pop[order]
    bpop = bpop[order]
    return pop, bpop

def bw_optimizer(scp, agent, iter_per_bs=5,  pp=0.6, cr=0.44, pm=0.4,npop=10, max_iter=50):

    # Inicializar población
    bounds = (-10,10)
    pop = inicializar_BWO(scp, npop, bounds)
    bpop = np.array(standar_bin(S_2(pop)))       # Binarizo S2 Standar

    for i in range(0, npop):
        bpop[i,:] = reparaSimple(scp, bpop[i,:])
    
    inicio = time.time()

    # Evaluar población
    pop, bpop = ordenar_fitness(scp, pop, bpop)

    gbest = bpop[0,:]
    fit_gbest = scp_ut.fitness(scp, gbest)
    fit_best = fit_gbest
    fit_worst = scp_ut.fitness(scp, bpop[npop-1,:])

    # Inicializar variables
    nr = int(npop*pp)     # numero de reproducciones
    nm = int(npop*pm)     # numero de mutaciones
    ns = int(npop*cr)     # numero de sobrevivientes

    hist = []
    hist.append(fit_gbest)

    old_fit = fit_best
    new_fit = np.zeros(iter_per_bs)

    tf , bf = update_bs(agent)
    mean_fit = 0
    # Iteraciones
    for i in range(0, max_iter):
        mab_iter = i % iter_per_bs

        pop1  = pop[:nr,:]           # poblacion de reproductores
        bpop1 = bpop[:nr,:]         # poblacion de reproductores binaria
        pop2  = pop[npop-1,:]       # poblacion resultante de la reproducción
        bpop2 = bpop[npop-1,:]     # poblacion resultante de la reproducción binaria

        #pop3 = np.zeros(nvar)       # poblacion resultante de la mutación
        #bpop3 = np.zeros(nvar)      # poblacion resultante de la mutación binaria

        #new_fit[mab_iter] = fit_gbest

        final = time.time()
        iter_time = final - inicio
        if i % 1 == 0:
            print('Iteración: ', i, '   Mejor Resultado: ', fit_best, "   Peor Resultado: ", fit_worst, " BS: ", tf, bf)
            print("Tiempo de iteración: ", iter_time)
        inicio = time.time()

        # Reproduciones
        for j in range(0, nr):
            aux = np.arange(0,pop1.shape[0])
            i1, i2 = np.random.choice(aux, replace=False) , np.random.choice(aux, replace=False)
            children, bchildren = crossover(scp, pop1, bpop1, i1, i2, tf, bf)
            
            # Canibalismo de padre
            pop1, bpop1 = father_cannibalism(scp, bpop1, pop1, i1, i2)
            
            # Canibalismo de hermanos
            children, bchildren = sibling_canibalism(scp, children, bchildren, cr)
            #exit(0)
            # ordenar por fitness
            # children, bchildren = ordenar_fitness(scp, children, bchildren)

            pop2 = np.vstack((pop2, children))
            bpop2 = np.vstack((bpop2, bchildren))
            

        # Mutaciones
        pop3, bpop3 = mutation(scp, pop2, bpop2, nm, tf, bf)

        pop2 = np.vstack((pop2, pop3))
        bpop2 = np.vstack((bpop2, bpop3))

        # ordenar por fitness
        pop2, bpop2 = ordenar_fitness(scp, pop2, bpop2)

        # Selección de mejores resultados
        pop = pop2[:npop,:]
        bpop = bpop2[:npop,:]

        # Mejor solución
        best = bpop[0,:]
        #  
        fit_best = scp_ut.fitness(scp, best)

        worst = bpop[npop-1,:]
        fit_worst = scp_ut.fitness(scp, worst)

        if fit_best < fit_gbest:
            gbest = best
            fit_gbest = scp_ut.fitness(scp, gbest)

        hist.append(fit_gbest)

        # Calculo reward
        new_fit[mab_iter] = fit_best

        # Actualizo Bin Scheme
        if mab_iter == iter_per_bs-1:
            mean_fit = np.mean(new_fit)

            agent.update_rw(old_fit, mean_fit)
            tf, bf = update_bs(agent)

            mean_fit = 0
            old_fit = np.mean(new_fit)

    return scp_ut.fitness(scp, gbest), gbest, agent

def main():
    # Lectura de archivo
    scp = scp_ut.read_scp('scp41.txt')

    # Inicializar entorno
    env = MAB.Env()

    # Inicializar agente
    agent = MAB.Agent(env, policy="ucb", arg=2)

    fit_gbest, gbest, agent = bw_optimizer(scp, agent, iter_per_bs=1, pp=0.6, cr=0.44, pm=0.4 ,npop=50, max_iter=200)

    print("fitness:", fit_gbest)
    print("gbest:", gbest)
    print("agent:", agent.memory["arms_played"])
    print(agent.memory["mean_reward"])
    print(agent.memory["rewards_hist"])


if __name__ == '__main__':
    main()


