import numpy as np
import random

## Bin scheme
def S_1(sol):
    trans = np.divide(1, ( 1 + np.exp(-2*sol) ) )
    return trans

def standar_bin(trans):
    matrixRand = np.random.uniform(low=0.0,high=1.0,size=trans.shape)
    bin = np.greater(trans,matrixRand).astype(int)
    return bin

def bin_scheme(sol):
    trans = S_1(sol)
    bin = standar_bin(trans)
    return np.array(bin)

class Env(object):
    def __init__(self ):
        transfer = ["S1", "S2", "S3", "S4", "V1", "V2", "V3", "V4"]
        binary = ["Standard", "Complement", "Elitist", "Static"]#, "ElitistRoulette"]
        self.k_arms = [(x,y) for x in transfer for y in binary]

    def reward_gen(self, old_fit, new_fit):
        rw = (old_fit - new_fit)/old_fit * 100
        return rw

    

class Agent(object):
    def __init__(self, env, policy="random", arg=0.1, exploration=True):
        self.policy = policy
        self.env = env
        self.arg        = arg
        self.memory     = {}
        self.memory["arms_played"] = []
        self.memory["mean_reward"] = []
        self.memory["rewards_hist"]= []
        self.exploration = exploration

    def update_rw(self, old, new):
        rw = self.env.reward_gen(old, new)
        h = self.memory["rewards_hist"]
        self.memory["rewards_hist"].append(rw)
        self.memory["mean_reward"].append(sum(h)/len(h))

    def select_arm(self):

        if self.policy == "random":
            arm = self.random()
            self.memory["arms_played"].append(arm)
            return arm
        elif self.policy == "epsilon":
            arm = self.epsilon()
            self.memory["arms_played"].append(arm)
            return arm
        elif self.policy == "ucb":
            arm = self.ucb()
            self.memory["arms_played"].append(arm)
            return arm
        else:
            print("Error: Policy not found")
            return None

    def mean_reward_per_arm(self, ucb=False):
        sum_rw = np.zeros(len(self.env.k_arms))
        cont_rw = np.zeros(len(self.env.k_arms))
        iters = len(self.memory["arms_played"])

        for i in range(iters):
            a = self.memory["arms_played"][i]       # i: Posición brazo seleccionado
            j = self.env.k_arms.index(a)            # j: Posición del brazo en el entorno
            sum_rw[j] += self.memory["rewards_hist"][i]
            cont_rw[j] += 1
        mean = np.divide(sum_rw, cont_rw, out=np.zeros_like(sum_rw), where=cont_rw!=0) # Si cont_rw == 0, devuelve 0
        if ucb:
            l = np.full(len(self.env.k_arms), np.log(iters))
            div = np.true_divide(l, cont_rw, out=np.zeros_like(l), where=cont_rw!=0)
            #div[div == np.NaN] = 100
            mean = mean + self.arg * np.sqrt(div)
        return mean

    def random(self):
        arm = random.choice(self.env.k_arms)  # Selecciono un brazo aleatorio
        return arm

    def epsilon(self):

        Q = self.mean_reward_per_arm()
        print(Q)

        if np.random.rand() <= self.arg or len(self.memory["arms_played"]) < 3:
            arm = random.choice(self.env.k_arms)
        else:
            index = np.where(Q == Q.max())  # lista de valores iguales al máximo
            print(index)
            arm = self.env.k_arms[np.random.choice(np.where(Q == Q.max())[0])]
        
        return arm

    def ucb(self):

        Q = self.mean_reward_per_arm(ucb = True)
        print(Q)
        
        if len(self.memory["arms_played"]) < len(self.env.k_arms):
            arm = random.choice(self.env.k_arms)
        else:
            arm = arm = self.env.k_arms[np.random.choice(np.where(Q == Q.max())[0])]

        return arm

        